"""
This module exposes components dedicated to the global configuration of
Eradiate.
"""

import lazy_loader

__getattr__, __dir__, __all__ = lazy_loader.attach_stub(__name__, __file__)

del lazy_loader
